package cd.casic.ci.process.api.process.pojo.app;

import cd.casic.ci.common.pipeline.constant.StringConstant;
import cd.casic.ci.common.pipeline.container.Container;
import cd.casic.ci.common.pipeline.container.Stage;
import cd.casic.ci.common.pipeline.enums.BuildStatus;
import cd.casic.ci.common.pipeline.enums.ChannelCode;
import cd.casic.ci.common.pipeline.enums.StartType;
import cd.casic.ci.common.pipeline.pojo.BuildNo;
import cd.casic.ci.common.pipeline.pojo.BuildParameters;
import cd.casic.ci.common.pipeline.pojo.element.Element;
import cd.casic.ci.common.pipeline.pojo.element.trigger.enums.CodeType;
import cd.casic.ci.common.pipeline.pojo.setting.PipelineRunLockType;
import cd.casic.ci.common.pipeline.pojo.setting.PipelineSetting;
import cd.casic.ci.common.pipeline.utils.CascadePropertyUtils;
import cd.casic.ci.common.pipeline.utils.EnvUtils;
import cd.casic.ci.event.enums.ActionType;
import cd.casic.ci.process.api.process.pojo.code.WebhookInfo;
import cd.casic.ci.process.api.process.utils.DependOnUtils;
import cd.casic.ci.process.api.process.utils.PipelineVarUtil;
import cd.casic.framework.commons.util.watcher.Watcher;
import lombok.AllArgsConstructor;
import lombok.Data;
import lombok.extern.slf4j.Slf4j;

import java.time.LocalDateTime;
import java.util.*;

import static cd.casic.ci.common.pipeline.constant.PipelineConstants.PIPELINE_GIT_EVENT_URL;
import static cd.casic.ci.common.pipeline.constant.WebhookConstant.*;
import static cd.casic.ci.process.api.process.utils.Constants.*;
import static cd.casic.ci.process.api.process.utils.PipelineVarUtil.CONTEXT_PREFIX;

/**
 * @author by mianbin
 * @Classname StartBuildContext
 * @Description TODO
 * @Date 2025/4/1 16:23
 */
@Data
@Slf4j
@AllArgsConstructor
public class StartBuildContext {
    private static final int MAX_LENGTH = 255;
    private static final int DELTA = 16;

    private final LocalDateTime now;
    private final String projectId;
    private final String pipelineId;
    private final String buildId;
    private final int resourceVersion;
    private final String versionName;
    private final String yamlVersion;
    private final ActionType actionType;
    private final int executeCount;
    private final boolean stageRetry;
    private final String retryStartTaskId;
    private String firstTaskId;
    private int containerSeq;
    private int taskCount;
    private final String userId;
    private final String triggerUser;
    private final StartType startType;
    private final String parentBuildId;
    private final String parentTaskId;
    private final ChannelCode channelCode;
    private final boolean retryFailedContainer;
    private boolean needUpdateStage;
    private final boolean skipFailedTask;
    private final Map<String, String> variables;
    private final BuildStatus startBuildStatus;
    private final WebhookInfo webhookInfo;
    private final String buildMsg;
    private final List<String> triggerReviewers;
    private final Map<String, BuildParameters> pipelineParamMap;
    private final List<BuildParameters> buildParameters;
    private final String concurrencyGroup;
    private final PipelineSetting pipelineSetting;
    private String buildNumAlias;
    private int buildNum;
    private BuildNo.BuildNoType buildNoType;
    private Integer currentBuildNo;
    private final boolean debug;
    private final String debugModelStr;
    private final Watcher watcher;

    public StartBuildContext(LocalDateTime now, String projectId, String pipelineId, String buildId, int resourceVersion, String versionName, String yamlVersion, ActionType actionType, int executeCount, boolean stageRetry, String retryStartTaskId, String firstTaskId, int containerSeq, int taskCount, String userId, String triggerUser, StartType startType, String parentBuildId, String parentTaskId, ChannelCode channelCode, boolean retryFailedContainer, boolean needUpdateStage, boolean skipFailedTask, Map<String, String> variables, BuildStatus startBuildStatus, WebhookInfo webhookInfo, String buildMsg, List<String> triggerReviewers, Map<String, BuildParameters> pipelineParamMap, List<BuildParameters> buildParameters, String concurrencyGroup, PipelineSetting pipelineSetting, String buildNumAlias, int buildNum, BuildNo.BuildNoType buildNoType, Integer currentBuildNo, boolean debug, String debugModelStr) {
        this.now = now != null ? now : LocalDateTime.now();
        this.projectId = projectId;
        this.pipelineId = pipelineId;
        this.buildId = buildId;
        this.resourceVersion = resourceVersion;
        this.versionName = versionName;
        this.yamlVersion = yamlVersion;
        this.actionType = actionType;
        this.executeCount = executeCount;
        this.stageRetry = stageRetry;
        this.retryStartTaskId = retryStartTaskId;
        this.firstTaskId = firstTaskId;
        this.containerSeq = containerSeq;
        this.taskCount = taskCount;
        this.userId = userId;
        this.triggerUser = triggerUser;
        this.startType = startType;
        this.parentBuildId = parentBuildId;
        this.parentTaskId = parentTaskId;
        this.channelCode = channelCode;
        this.retryFailedContainer = retryFailedContainer;
        this.needUpdateStage = needUpdateStage;
        this.skipFailedTask = skipFailedTask;
        this.variables = variables;
        this.startBuildStatus = startBuildStatus;
        this.webhookInfo = webhookInfo;
        this.buildMsg = buildMsg;
        this.triggerReviewers = triggerReviewers;
        this.pipelineParamMap = pipelineParamMap;
        this.buildParameters = buildParameters;
        this.concurrencyGroup = concurrencyGroup;
        this.pipelineSetting = pipelineSetting;
        this.buildNumAlias = buildNumAlias;
        this.buildNum = buildNum;
        this.buildNoType = buildNoType;
        this.currentBuildNo = currentBuildNo;
        this.debug = debug;
        this.debugModelStr = debugModelStr;
        this.watcher = new Watcher("startBuild-" + buildId);
    }

    /**
     * 检查Stage是否属于失败重试[stageRetry]时，当前[stage]是否需要跳过
     */
    public boolean needSkipWhenStageFailRetry(Stage stage) {
        if (needRerunStage(stage)) {
            return false;
        } else if (!stageRetry) {
            return false;
        } else {
            return BuildStatus.parse(stage.getStatus()).isFinish();
        }
    }

    public boolean needSkipContainerWhenFailRetry(Stage stage, Container container) {
        BuildStatus containerStatus = BuildStatus.parse(container.getStatus());
        if (needRerunStage(stage)) {
            return false;
        } else if (!containerStatus.isFailure() && !containerStatus.isCancel()) {
            return false;
        } else {
            return !retryStartTaskId.isEmpty();
        }
    }

    public boolean needSkipTaskWhenRetry(Stage stage, Container container, String taskId) {
        if (stage.isFinally()) {
            return false;
        } else if (stage.getId().equals(retryStartTaskId)) {
            return false;
        } else if (retryStartTaskId.isEmpty()) {
            return false;
        } else if (isRetryDependOnContainer(container)) {
            return false;
        } else {
            return !retryStartTaskId.equals(taskId);
        }
    }

    public boolean inSkipStage(Stage stage, Element atom) {
        if (skipFailedTask && retryStartTaskId.equals(atom.getId())) {
            return true;
        } else {
            return skipFailedTask && (stage.getId().equals(retryStartTaskId) && BuildStatus.parse(atom.getStatus()).isFailure());
        }
    }

    /**
     * 是否是要重试的失败容器
     */
    public boolean isRetryFailedContainer(Stage stage, Container container) {
        if (stage.isFinally()) {
            if (stage.getId().equals(retryStartTaskId)) {
                return retryFailedContainer && BuildStatus.parse(container.getStatus()).isSuccess();
            } else {
                return false;
            }
        } else if (isRetryDependOnContainer(container)) {
            return false;
        } else {
            return retryFailedContainer && BuildStatus.parse(container.getStatus()).isSuccess();
        }
    }

    // 失败重试,跳过的dependOn容器也应该被执行
    private boolean isRetryDependOnContainer(Container container) {
        return DependOnUtils.enableDependOn(container) && BuildStatus.parse(container.getStatus()) == BuildStatus.SKIP;
    }

    public boolean needRerunStage(Stage stage) {
        return stage.isFinally() || retryStartTaskId == null || stage.getId().equals(retryStartTaskId);
    }

    public boolean needRerunTask(Stage stage, Container container) {
        return needRerunStage(stage) || isRetryDependOnContainer(container);
    }

    public static StartBuildContext init(String projectId, String pipelineId, String buildId, int resourceVersion, String versionName, String yamlVersion, String modelStr, boolean debug, PipelineSetting pipelineSetting, List<String> realStartParamKeys, Map<String, BuildParameters> pipelineParamMap, Map<String, BuildParameters> webHookStartParam, List<String> triggerReviewers, Integer currentBuildNo) {
        List<BuildParameters> buildParam = genOriginStartParamsList(realStartParamKeys, pipelineParamMap);
        Map<String, String> params = new HashMap<>();
        for (BuildParameters param : pipelineParamMap.values()) {
            params.put(param.getKey(), param.getValue().toString());
        }
        String retryStartTaskId = params.get(PIPELINE_RETRY_START_TASK_ID);

        ActionType actionType;
        int executeCount;
        boolean isStageRetry;
        if (params.get(PIPELINE_RETRY_COUNT) != null) {
            int count;
            try {
                count = Integer.parseInt(params.get(PIPELINE_RETRY_COUNT).trim());
                count = Math.max(count, 0);
            } catch (NumberFormatException ignored) {
                count = 0;
            }
            actionType = ActionType.RETRY;
            executeCount = count + 1;
            isStageRetry = retryStartTaskId != null && retryStartTaskId.startsWith("stage-");
        } else {
            actionType = ActionType.START;
            executeCount = 1;
            isStageRetry = false;
        }
        ChannelCode channelCode = params.get(PIPELINE_START_CHANNEL) != null ? ChannelCode.valueOf(params.get(PIPELINE_START_CHANNEL)) : ChannelCode.BS;
        boolean retryFailedContainer = params.get(PIPELINE_RETRY_ALL_FAILED_CONTAINER) != null ? Boolean.parseBoolean(params.get(PIPELINE_RETRY_ALL_FAILED_CONTAINER)) : false;
        boolean skipFailedTask = params.get(PIPELINE_SKIP_FAILED_TASK) != null ? Boolean.parseBoolean(params.get(PIPELINE_SKIP_FAILED_TASK)) : false;
        String buildMsg = params.get(PIPELINE_BUILD_MSG) != null ? StringConstant.coerceAtMaxLength(params.get(PIPELINE_BUILD_MSG), MAX_LENGTH) : null;

        String concurrencyGroup = null;
        if (pipelineSetting != null && pipelineSetting.getRunLockType() == PipelineRunLockType.GROUP_LOCK && pipelineSetting.getConcurrencyGroup() != null) {
            Map<String, String> webhookParam = new HashMap<>();
            for (BuildParameters param : webHookStartParam.values()) {
                webhookParam.put(param.getKey(), param.getValue().toString());
            }
            concurrencyGroup = EnvUtils.parseEnv(pipelineSetting.getConcurrencyGroup(), PipelineVarUtil.fillContextVarMap(webhookParam));
            log.info("[$pipelineId]|[$buildId]|ConcurrencyGroup=$concurrencyGroup");
        }

        BuildStatus startBuildStatus = triggerReviewers == null || triggerReviewers.isEmpty() ? BuildStatus.QUEUE : BuildStatus.TRIGGER_REVIEWING;

        return new StartBuildContext(
                LocalDateTime.now(),
                projectId,
                pipelineId,
                buildId,
                resourceVersion,
                versionName,
                yamlVersion,
                actionType,
                executeCount,
                isStageRetry,
                retryStartTaskId,
                params.getOrDefault(PIPELINE_START_TASK_ID, ""),
                0,
                0,
                params.get(PIPELINE_START_USER_ID),
                params.get(PIPELINE_START_USER_NAME),
                StartType.valueOf(params.get(PIPELINE_START_TYPE)),
                params.get(PIPELINE_START_PARENT_BUILD_ID),
                params.get(PIPELINE_START_PARENT_BUILD_TASK_ID),
                channelCode,
                retryFailedContainer,
                false,
                skipFailedTask,
                params,
                startBuildStatus,
                getWebhookInfo(params),
                buildMsg,
                triggerReviewers,
                pipelineParamMap,
                buildParam,
                concurrencyGroup,
                pipelineSetting,
                null,
                1,
                null,
                currentBuildNo,
                debug,
                modelStr
        );
    }


    private static WebhookInfo getWebhookInfo(Map<String, String> params) {
        List<String> startTypes = Arrays.asList(StartType.WEB_HOOK.name(), StartType.PIPELINE.name(), StartType.SERVICE.name(), StartType.REMOTE.name());
        String startType = params.get(PIPELINE_START_TYPE);
        if (!startTypes.contains(startType)) {
            return null;
        }
        return new WebhookInfo(
                supportCustomMaterials(startType) ? startType : params.get(REPO_WEBHOOK_REPO_TYPE),
                params.get(REPO_WEBHOOK_REPO_NAME),
                params.get(PIPELINE_WEBHOOK_COMMIT_MESSAGE),
                params.get(REPO_WEBHOOK_REPO_URL),
                params.get(PIPELINE_WEBHOOK_TYPE),
                params.get(PIPELINE_WEBHOOK_BRANCH),
                params.get(REPO_WEBHOOK_REPO_ALIAS_NAME),
                params.get(PIPELINE_WEBHOOK_TYPE).equals(CodeType.GIT.name()) ? params.get(REPO_GIT_WEBHOOK_EVENT_TYPE) : params.get(PIPELINE_WEBHOOK_EVENT_TYPE),
                params.get(PIPELINE_WEBHOOK_REVISION),
                params.get(PIPELINE_WEBHOOK_REVISION),
                params.get(REPO_GIT_WEBHOOK_MR_MERGE_COMMIT_SHA),
                params.get(REPO_GIT_WEBHOOK_MR_SOURCE_BRANCH),
                params.get(REPO_GIT_WEBHOOK_MR_ID),
                params.get(REPO_GIT_WEBHOOK_MR_NUMBER),
                params.get(REPO_GIT_WEBHOOK_MR_URL),
                params.get(REPO_WEBHOOK_REPO_AUTH_USER),
                params.get(REPO_GIT_WEBHOOK_TAG_NAME),
                params.get(REPO_GIT_WEBHOOK_ISSUE_IID),
                params.get(REPO_GIT_WEBHOOK_NOTE_ID),
                params.get(REPO_GIT_WEBHOOK_REVIEW_ID),
                params.get(PIPELINE_START_PARENT_PROJECT_ID),
                params.get(PIPELINE_START_PARENT_PIPELINE_ID),
                params.get(PIPELINE_START_PARENT_PIPELINE_NAME),
                params.get(PIPELINE_START_PARENT_BUILD_ID),
                params.get(PIPELINE_START_PARENT_BUILD_NUM),
                supportCustomMaterials(startType) ? params.get(BK_CI_MATERIAL_URL) : params.get(PIPELINE_GIT_EVENT_URL),
                params.get(BK_CI_MATERIAL_ID),
                params.get(BK_CI_MATERIAL_NAME)
        );
    }

    /**
     * 是否支持自定义触发材料
     */
    private static boolean supportCustomMaterials(String startType) {
        return startType != null && (startType.equals(StartType.REMOTE.name()) || startType.equals(StartType.SERVICE.name()));
    }

    public static StartBuildContext init4SendBuildStartEvent(String userId, String projectId, String pipelineId, String buildId, int resourceVersion, String versionName, ActionType actionType, int executeCount, String firstTaskId, StartType startType, BuildStatus startBuildStatus, boolean debug) {
        return new StartBuildContext(
                LocalDateTime.now(),
                projectId,
                pipelineId,
                buildId,
                resourceVersion,
                versionName,
                null,
                actionType,
                executeCount,
                false,
                null,
                firstTaskId,
                0,
                0,
                userId,
                "",
                startType,
                "",
                "",
                ChannelCode.BS,
                false,
                false,
                false,
                Collections.emptyMap(),
                startBuildStatus,
                null,
                null,
                null,
                new HashMap<>(),
                new ArrayList<>(),
                null,
                null,
                null,
                1,
                null,
                null,
                debug,
                null
        );
    }

    /**
     * 根据[realStartParamKeys]启动参数Key列表读取[pipelineParamMap]参数值来生成流水线启动变量列表，不包含其他
     */
    private static ArrayList<BuildParameters> genOriginStartParamsList(List<String> realStartParamKeys, Map<String, BuildParameters> pipelineParamMap) {
        ArrayList<BuildParameters> originStartParams = new ArrayList<>(realStartParamKeys.size() + DELTA);
        HashMap<String, BuildParameters> originStartContexts = new HashMap<>(realStartParamKeys.size());
        for (String key : realStartParamKeys) {
            BuildParameters param = pipelineParamMap.get(key);
            if (param != null) {
                if (CascadePropertyUtils.supportCascadeParam(param.getValueType())) {
                    originStartParams.addAll(fillCascadeParam(param, originStartContexts));
                } else {
                    originStartParams.add(param);
                    fillContextPrefix(param, originStartContexts);
                }
            }
        }
        pipelineParamMap.putAll(originStartContexts);

        if (pipelineParamMap.containsKey(BUILD_NO)) {
            originStartParams.add(pipelineParamMap.get(BUILD_NO));
        }
        if (pipelineParamMap.containsKey(PIPELINE_BUILD_MSG)) {
            originStartParams.add(pipelineParamMap.get(PIPELINE_BUILD_MSG));
        }
        if (pipelineParamMap.containsKey(PIPELINE_RETRY_COUNT)) {
            originStartParams.add(pipelineParamMap.get(PIPELINE_RETRY_COUNT));
        }

        return originStartParams;
    }

    private static void fillContextPrefix(BuildParameters param, HashMap<String, BuildParameters> originStartContexts) {
        String key = param.getKey();
        if (key.startsWith(CONTEXT_PREFIX)) {
            originStartContexts.put(key, param);
        } else {
            String ctxKey = CONTEXT_PREFIX + key;
            originStartContexts.put(ctxKey, param);
        }
    }

    /**
     * 根据原始值，填充级联参数
     * xxx = {"repo-name": "xxx/xxx","branch":"master"}
     * xxx.repo-name = xxx/xxx
     * xxx.branch = master
     */
    private static List<BuildParameters> fillCascadeParam(BuildParameters param, HashMap<String, BuildParameters> originStartContexts) {
        List<BuildParameters> originStartParams = new ArrayList<>();
        String key = param.getKey();
        Object paramValue = CascadePropertyUtils.parseDefaultValue(key, param.getValue(), param.getValueType());
        BuildParameters cascadeParam = new BuildParameters(
                param.getKey(),
                paramValue,
                param.getValueType(),
                param.getDesc(),
                param.getDefaultValue()
        );
        originStartParams.add(cascadeParam);
        // 填充下级参数的[variables.]
        fillContextPrefix(cascadeParam, originStartContexts);
        Map<String, String> cascadeVariableKeyMap = CascadePropertyUtils.getCascadeVariableKeyMap(key, param.getValueType());
        for (Map.Entry<String, String> entry : cascadeVariableKeyMap.entrySet()) {
            String subKey = entry.getKey();
            String paramKey = entry.getValue();
            BuildParameters subParam = new BuildParameters(
                    paramKey,
                    paramValue != null && ((Map<?, ?>) paramValue).containsKey(subKey) ? ((Map<?, ?>) paramValue).get(subKey) : "",
                    param.getValueType(),
                    param.getDesc(),
                    param.getDefaultValue()
            );
            // 填充下级参数的[variables.]
            fillContextPrefix(subParam, originStartContexts);
        }
        return originStartParams;
    }
}
